@@ -39,8 +39,6 @@
 #define FONT_INDEX_FUDGE 1
 #define CHAR_WIDTH_FUDGE 1
 
-#define DPY_FUDGE 1
-
 /* libXft will accept runs up to 1024 glyphs before allocating a temporary
  * array. However, setting this to a large value can cause dramatic slow-downs
  * for some xservers (notably fglrx), see bug 410534.
@@ -70,7 +68,6 @@ struct _vte_xft_data {
 	Visual *visual;
 	Colormap colormap;
 	XftDraw *draw;
-	GC gc;
 	GdkColor color;
 	guint16 opacity;
 	GdkPixmap *pixmap;
@@ -344,13 +341,16 @@ static void
 _vte_xft_create (struct _vte_draw *draw, GtkWidget *widget)
 {
 	struct _vte_xft_data *data;
+
 	data = g_slice_new0 (struct _vte_xft_data);
 	draw->impl_data = data;
-	data->drawable = -1;
-	data->colormap = -1;
+
 	data->opacity = 0xffff;
+
 	data->xpixmap = -1;
 	data->pixmapw = data->pixmaph = -1;
+
+	data->drawable = -1;
 }
 
 static void
@@ -381,9 +381,6 @@ _vte_xft_destroy (struct _vte_draw *draw)
 	if (data->draw != NULL) {
 		XftDrawDestroy (data->draw);
 	}
-	if (data->gc != NULL) {
-		XFreeGC (data->display, data->gc);
-	}
 	g_slice_free (struct _vte_xft_data, data);
 }
 
@@ -402,8 +399,6 @@ _vte_xft_get_colormap (struct _vte_draw *draw)
 static void
 _vte_xft_start (struct _vte_draw *draw)
 {
-	GdkVisual *gvisual;
-	GdkColormap *gcolormap;
 	GdkDrawable *drawable;
 	GPtrArray *locked_fonts;
 	guint i;
@@ -411,33 +406,31 @@ _vte_xft_start (struct _vte_draw *draw)
 	struct _vte_xft_data *data;
 	data = (struct _vte_xft_data*) draw->impl_data;
 
+	gdk_error_trap_push ();
+
 	gdk_window_get_internal_paint_info (draw->widget->window,
 					   &drawable,
 					   &data->x_offs,
 					   &data->y_offs);
+	if (data->drawable != gdk_x11_drawable_get_xid (drawable)) {
+		GdkVisual *gvisual;
+		GdkColormap *gcolormap;
 
-	data->display = gdk_x11_drawable_get_xdisplay (drawable);
-	data->drawable = gdk_x11_drawable_get_xid (drawable);
-	gvisual = gdk_drawable_get_visual (drawable);
-	data->visual = gdk_x11_visual_get_xvisual (gvisual);
-	gcolormap = gdk_drawable_get_colormap (drawable);
-	data->colormap = gdk_x11_colormap_get_xcolormap (gcolormap);
-
+		if (data->draw != NULL) {
+			XftDrawDestroy (data->draw);
+		}
+		data->display = gdk_x11_drawable_get_xdisplay (drawable);
+		data->drawable = gdk_x11_drawable_get_xid (drawable);
+		gvisual = gdk_drawable_get_visual (drawable);
+		data->visual = gdk_x11_visual_get_xvisual (gvisual);
+		gcolormap = gdk_drawable_get_colormap (drawable);
+		data->colormap = gdk_x11_colormap_get_xcolormap (gcolormap);
+		data->draw = XftDrawCreate (data->display, data->drawable,
+				data->visual, data->colormap);
+	}
 	g_assert (data->display == data->font->display);
 
-	gdk_error_trap_push ();
-
-	if (data->draw != NULL) {
-		XftDrawDestroy (data->draw);
-	}
-	data->draw = XftDrawCreate (data->display, data->drawable,
-				   data->visual, data->colormap);
-	if (data->gc != NULL) {
-		XFreeGC (data->display, data->gc);
-	}
-	data->gc = XCreateGC (data->display, data->drawable, 0, NULL);
-
-	locked_fonts = data->locked_fonts [(++data->cur_locked_fonts)&1];
+	locked_fonts = data->locked_fonts [++data->cur_locked_fonts&1];
 	if (locked_fonts != NULL) {
 		guint cnt=0;
 		for (i = 1; i < locked_fonts->len; i++) {
@@ -457,16 +450,6 @@ _vte_xft_end (struct _vte_draw *draw)
 	struct _vte_xft_data *data;
 
 	data = (struct _vte_xft_data*) draw->impl_data;
-	if (data->draw != NULL) {
-		XftDrawDestroy (data->draw);
-		data->draw = NULL;
-	}
-	if (data->gc != NULL) {
-		XFreeGC (data->display, data->gc);
-		data->gc = NULL;
-	}
-	data->drawable = -1;
-	data->x_offs = data->y_offs = 0;
 
 	gdk_error_trap_pop ();
 }
@@ -479,6 +462,9 @@ _vte_xft_set_background_color (struct _vte_draw *draw, GdkColor *color,
 	data = (struct _vte_xft_data*) draw->impl_data;
 	data->color = *color;
 	data->opacity = opacity;
+
+	draw->requires_clear = opacity != 0xffff
+		|| (data->pixmapw > 0 && data->pixmaph > 0);
 }
 
 static void
@@ -505,13 +491,13 @@ _vte_xft_set_background_image (struct _vte_draw *draw,
 	if (data->pixmap != NULL) {
 		g_object_unref (data->pixmap);
 	}
-	draw->has_background_image = FALSE;
+	draw->requires_clear = data->opacity != 0xffff;
 	data->pixmap = NULL;
 	if (pixmap != NULL) {
 		data->pixmap = pixmap;
 		data->xpixmap = gdk_x11_drawable_get_xid (pixmap);
 		gdk_drawable_get_size (pixmap, &data->pixmapw, &data->pixmaph);
-		draw->has_background_image =
+		draw->requires_clear |=
 			data->pixmapw > 0 && data->pixmaph > 0;
 	}
 }
@@ -521,25 +507,24 @@ _vte_xft_clip (struct _vte_draw *draw,
 		GdkRegion *region)
 {
 	struct _vte_xft_data *data = draw->impl_data;
+	XRectangle stack_rect[16];
+	XRectangle *xrect;
 	GdkRectangle *rect;
 	gint i, n;
 
 	gdk_region_get_rectangles (region, &rect, &n);
-	if (n>0) {
-		XRectangle *xrect = g_new (XRectangle, n);
-		for (i = 0; i < n; i++) {
-			/* we include the offset here as XftDrawSetClipRectangles () has a
-			 * byte-sex bug in its offset parameters. Bug 403159.
-			 */
-			xrect[i].x = rect[i].x - data->x_offs;
-			xrect[i].y = rect[i].y - data->y_offs;
-			xrect[i].width = rect[i].width;
-			xrect[i].height = rect[i].height;
-		}
-		XftDrawSetClipRectangles (data->draw,
-				0, 0, xrect, n);
-		g_free (xrect);
+	xrect = n > (gint) G_N_ELEMENTS (stack_rect) ?
+		g_new (XRectangle, n) :
+		stack_rect;
+	for (i = 0; i < n; i++) {
+		xrect[i].x = rect[i].x - data->x_offs;
+		xrect[i].y = rect[i].y - data->y_offs;
+		xrect[i].width = rect[i].width;
+		xrect[i].height = rect[i].height;
 	}
+	XftDrawSetClipRectangles (data->draw, 0, 0, xrect, n);
+	if (xrect != stack_rect)
+		g_free (xrect);
 	g_free (rect);
 }
 
@@ -551,6 +536,7 @@ _vte_xft_clear (struct _vte_draw *draw,
 	XRenderColor rcolor;
 	XftColor ftcolor;
 	gint h, w, txstop, tystop, sx, sy, tx, ty;
+	GC gc;
 
 	data = (struct _vte_xft_data*) draw->impl_data;
 
@@ -580,6 +566,7 @@ _vte_xft_clear (struct _vte_draw *draw,
 	tystop = y + height;
 
 	/* Flood fill. */
+	gc = XCreateGC (data->display, data->drawable, 0, NULL);
 	sy = (data->scrolly + y) % data->pixmaph;
 	while (ty < tystop) {
 		h = MIN (data->pixmaph - sy, tystop - ty);
@@ -590,7 +577,7 @@ _vte_xft_clear (struct _vte_draw *draw,
 			XCopyArea (data->display,
 				  data->xpixmap,
 				  data->drawable,
-				  data->gc,
+				  gc,
 				  sx, sy,
 				  w, h,
 				  tx - data->x_offs, ty - data->y_offs);
@@ -600,6 +587,7 @@ _vte_xft_clear (struct _vte_draw *draw,
 		ty += h;
 		sy = 0;
 	}
+	XFreeGC (data->display, gc);
 }
 
 static GPtrArray *
@@ -791,7 +779,7 @@ _vte_xft_draw_text (struct _vte_draw *draw,
 	XftColor ftcolor;
 	struct _vte_xft_data *data;
 	gsize i, j;
-	gint width, y_off, x_off;
+	gint width, y_off, x_off, char_width;
 	XftFont *font, *ft;
 	GPtrArray *locked_fonts;
 
@@ -831,17 +819,21 @@ _vte_xft_draw_text (struct _vte_draw *draw,
 	 * haven't pinned down yet." */
 	x_off = -data->x_offs;
 	y_off = draw->ascent - data->y_offs;
+	char_width = draw->width;
 	do {
 		j = 0;
 		do {
+			gint next_x;
+
 			glyphs[j].glyph = XftCharIndex (data->display,
 					font, requests[i].c);
 			glyphs[j].x = requests[i].x + x_off;
+			next_x = requests[i].x + requests[i].columns*char_width;
 			width = _vte_xft_char_width (data->font,
 					font, requests[i].c, requests[i].columns);
 			if (G_UNLIKELY (width != 0)) {
-				width = requests[i].columns * draw->width - width;
-				width = CLAMP (width / 2, 0, draw->width);
+				width = requests[i].columns*char_width - width;
+				width = CLAMP (width / 2, 0, char_width);
 				glyphs[j].x += width;
 			}
 			glyphs[j].y = requests[i].y + y_off;
@@ -857,7 +849,31 @@ _vte_xft_draw_text (struct _vte_draw *draw,
 				}
 				break;
 			}
-		} while (j < MAX_RUN_LENGTH && ft == font);
+			if (j == MAX_RUN_LENGTH || ft != font) {
+				break;
+			}
+
+			/* check to see if we've skipped over any spaces...
+			 * and reinsert them so as not to break the stream
+			 * unnecessarily - the blank space is less overhead
+			 * than starting a new sequence.
+			 */
+			if (requests[i].y + y_off == glyphs[j-1].y) {
+				while (next_x < requests[i].x) {
+					glyphs[j].glyph = XftCharIndex (
+							data->display,
+							font,
+							' ');
+					glyphs[j].x = next_x + x_off;
+					glyphs[j].y = glyphs[j-1].y;
+					if (++j == MAX_RUN_LENGTH) {
+						goto draw;
+					}
+					next_x += char_width;
+				}
+			}
+		} while (TRUE);
+draw:
 		XftDrawGlyphSpec (data->draw, &ftcolor, font, glyphs, j);
 		font = ft;
 	} while (i < n_requests);
